"""Utility function for the workcell manager."""

import shutil
from copy import deepcopy
from datetime import datetime
from pathlib import Path
from typing import Any, Optional

from fastapi import UploadFile
from madsci.client.event_client import default_logger
from madsci.common.types.auth_types import OwnershipInfo
from madsci.common.types.location_types import Location
from madsci.common.types.step_types import Step
from madsci.common.types.workcell_types import WorkcellDefinition
from madsci.common.types.workflow_types import (
    Workflow,
    WorkflowDefinition,
    WorkflowStatus,
)
from madsci.workcell_manager.redis_handler import WorkcellRedisHandler


def validate_node_names(workflow: Workflow, workcell: WorkcellDefinition) -> None:
    """
    Validates that the nodes in the workflow.step are in the workcell.nodes
    """
    for node_name in [step.node for step in workflow.steps]:
        if node_name not in workcell.nodes:
            raise ValueError(f"Node {node_name} not in Workcell {workcell.name}")


def validate_step(step: Step, state_handler: WorkcellRedisHandler) -> tuple[bool, str]:
    """Check if a step is valid based on the node's info"""
    if step.node in state_handler.get_all_nodes():
        node = state_handler.get_node(step.node)
        info = node.info
        if info is None:
            return (
                True,
                f"Node {step.node} didn't return proper about information, skipping validation",
            )
        if step.action in info.actions:
            action = info.actions[step.action]
            for action_arg in action.args.values():
                if action_arg.name not in step.args and action_arg.required:
                    return (
                        False,
                        f"Step '{step.name}': Node {step.node}'s action, '{step.action}', is missing arg '{action_arg.name}'",
                    )
                # TODO: Action arg type validation goes here
            for action_file in action.files:
                if action_file.name not in step.files and action_file.required:
                    return (
                        False,
                        f"Step '{step.name}': Node {step.node}'s action, '{step.action}', is missing file '{action_file.name}'",
                    )
            return True, f"Step '{step.name}': Validated successfully"

        return (
            False,
            f"Step '{step.name}': Node {step.node} has no action '{step.action}'",
        )
    return (
        False,
        f"Step '{step.name}': Node {step.node} is not defined in workcell",
    )


def create_workflow(
    workflow_def: WorkflowDefinition,
    workcell: WorkcellDefinition,
    state_handler: WorkcellRedisHandler,
    ownership_info: Optional[OwnershipInfo] = None,
    parameters: Optional[dict[str, Any]] = None,
) -> Workflow:
    """Pulls the workcell and builds a list of dictionary steps to be executed

    Parameters
    ----------
    workflow_def: WorkflowDefintion
        The workflow data file loaded in from the workflow yaml file

    workcell : Workcell
        The Workcell object stored in the database

    parameters: Dict
        The input to the workflow

    ownership_info: OwnershipInfo
        Information on the owner(s) of the workflow

    simulate: bool
        Whether or not to use real robots

    Returns
    -------
    steps: WorkflowRun
        a completely initialized workflow run
    """
    validate_node_names(workflow_def, workcell)
    wf_dict = workflow_def.model_dump()
    wf_dict.update(
        {
            "label": workflow_def.name,
            "ownership_info": ownership_info.model_dump(mode="json"),
            "parameter_values": parameters,
        }
    )
    wf = Workflow(**wf_dict)
    wf.step_definitions = workflow_def.steps
    steps = []
    for step in workflow_def.steps:
        working_step = deepcopy(step)
        replace_locations(workcell, working_step)
        valid, validation_string = validate_step(
            working_step, state_handler=state_handler
        )
        default_logger.log_info(validation_string)
        if not valid:
            raise ValueError(validation_string)
        steps.append(working_step)

    wf.steps = steps
    wf.submitted_time = datetime.now()
    return wf


def replace_locations(workcell: WorkcellDefinition, step: Step) -> None:
    """Replaces the location names with the location objects"""
    for location_arg, location_name_or_object in step.locations.items():
        if isinstance(location_name_or_object, Location):
            step.locations[location_arg] = location_name_or_object
        elif location_name_or_object in workcell.locations:
            step.locations[location_arg] = workcell.locations[location_name_or_object]
        else:
            raise ValueError(
                f"Location {location_name_or_object} not in Workcell {workcell.name}"
            )


def save_workflow_files(
    working_directory: str, workflow: Workflow, files: list[UploadFile]
) -> Workflow:
    """Saves the files to the workflow run directory,
    and updates the step files to point to the new location"""

    get_workflow_inputs_directory(
        workflow_id=workflow.workflow_id, working_directory=working_directory
    ).expanduser().mkdir(parents=True, exist_ok=True)
    if files:
        for file in files:
            file_path = (
                get_workflow_inputs_directory(
                    working_directory=working_directory,
                    workflow_id=workflow.workflow_id,
                )
                / file.filename
            ).expanduser()
            with Path.open(file_path, "wb") as f:
                f.write(file.file.read())
            for step in workflow.steps:
                for step_file_key, step_file_path in step.files.items():
                    if step_file_path == file.filename:
                        step.files[step_file_key] = str(file_path)
                        default_logger.log_info(
                            f"{step_file_key}: {file_path} ({step_file_path})"
                        )
    return workflow


def copy_workflow_files(
    working_directory: str, old_id: str, workflow: Workflow
) -> Workflow:
    """Saves the files to the workflow run directory,
    and updates the step files to point to the new location"""

    new = get_workflow_inputs_directory(
        workflow_id=workflow.workflow_id, working_directory=working_directory
    )
    old = get_workflow_inputs_directory(
        workflow_id=old_id, working_directory=working_directory
    )
    shutil.copytree(old, new)
    return workflow


def get_workflow_inputs_directory(
    workflow_id: Optional[str] = None, working_directory: Optional[str] = None
) -> Path:
    """returns a directory name for the workflows inputs"""
    return Path(working_directory).expanduser() / "Workflows" / workflow_id / "Inputs"


def cancel_workflow(wf: Workflow, state_handler: WorkcellRedisHandler) -> None:
    """Cancels the workflow run"""
    wf.status = WorkflowStatus.CANCELLED
    with state_handler.wc_state_lock():
        state_handler.set_workflow(wf)
    return wf


def cancel_active_workflows(state_handler: WorkcellRedisHandler) -> None:
    """Cancels all currently running workflow runs"""
    for wf in state_handler.get_all_workflows().values():
        if wf.status in [
            WorkflowStatus.RUNNING,
            WorkflowStatus.QUEUED,
            WorkflowStatus.IN_PROGRESS,
        ]:
            cancel_workflow(wf, state_handler=state_handler)



"""
State management for the WorkcellManager
"""

import warnings
from typing import Any, Callable, Optional, Union

import redis
from madsci.common.types.base_types import new_ulid_str
from madsci.common.types.node_types import Node, NodeDefinition
from madsci.common.types.workcell_types import WorkcellDefinition
from madsci.common.types.workflow_types import Workflow
from pottery import InefficientAccessWarning, RedisDict, RedisList, Redlock
from pydantic import ValidationError


class WorkcellRedisHandler:
    """
    Manages state for WEI, providing transactional access to reading and writing state with
    optimistic check-and-set and locking.
    """

    state_change_marker = "0"
    _redis_connection: Any = None

    def __init__(
        self,
        workcell_definition: WorkcellDefinition,
        redis_connection: Optional[Any] = None,
    ) -> None:
        """
        Initialize a StateManager for a given workcell.
        """
        self._workcell_name = workcell_definition.workcell_name
        self._redis_host = workcell_definition.config.redis_host
        self._redis_port = workcell_definition.config.redis_port
        self._redis_password = workcell_definition.config.redis_password
        self._redis_connection = redis_connection
        warnings.filterwarnings("ignore", category=InefficientAccessWarning)

    @property
    def _workcell_prefix(self) -> str:
        return f"workcell:{self._workcell_name}"

    @property
    def _redis_client(self) -> Any:
        """
        Returns a redis.Redis client, but only creates one connection.
        MyPy can't handle Redis object return types for some reason, so no type-hinting.
        """
        if self._redis_connection is None:
            self._redis_connection = redis.Redis(
                host=str(self._redis_host),
                port=int(self._redis_port),
                db=0,
                decode_responses=True,
                password=self._redis_password if self._redis_password else None,
            )
        return self._redis_connection

    @property
    def _workcell(self) -> RedisDict:
        return RedisDict(
            key=f"{self._workcell_prefix}:workcell", redis=self._redis_client
        )

    @property
    def _nodes(self) -> RedisDict:
        return RedisDict(key=f"{self._workcell_prefix}:nodes", redis=self._redis_client)

    @property
    def _workflow_queue(self) -> RedisList:
        return RedisList(
            key=f"{self._workcell_prefix}:workflow_queue", redis=self._redis_client
        )

    @property
    def _workflows(self) -> RedisDict:
        return RedisDict(
            key=f"{self._workcell_prefix}:workflows", redis=self._redis_client
        )

    def wc_state_lock(self) -> Redlock:
        """
        Gets a lock on the workcell's state. This should be called before any state updates are made,
        or where we don't want the state to be changing underneath us (i.e., in the engine).
        """
        return Redlock(
            key=f"{self._workcell_prefix}:state_lock",
            masters={self._redis_client},
            auto_release_time=60,
        )

    # *State Methods
    def get_state(self) -> dict[str, dict[Any, Any]]:
        """
        Return a dict containing the current state of the workcell.
        """
        return {
            "status": self.wc_status,
            "error": self.error,
            "nodes": self._nodes.to_dict(),
            "workflows": self._workflow_runs.to_dict(),
            "workcell": self._workcell.to_dict(),
            "paused": self.paused,
            "locked": self.locked,
            "shutdown": self.shutdown,
        }

    @property
    def error(self) -> str:
        """Latest error on the server"""
        return self._redis_client.get(f"{self._workcell_prefix}:error")

    @error.setter
    def error(self, value: str) -> None:
        """Add an error to the workcell's error deque"""
        if self.error != value:
            self.mark_state_changed()
        return self._redis_client.set(f"{self._workcell_prefix}:error", value)

    def clear_state(self, clear_workflows: bool = False) -> None:
        """
        Clears the state of the workcell, optionally leaving the locations state intact.
        """
        self._nodes.clear()
        if clear_workflows:
            self._workflows.clear()
        self.state_change_marker = "0"
        self.paused = False
        self.locked = False
        self.shutdown = False
        self.mark_state_changed()

    def mark_state_changed(self) -> int:
        """Marks the state as changed and returns the current state change counter"""
        return int(self._redis_client.incr(f"{self._workcell_prefix}:state_changed"))

    def has_state_changed(self) -> bool:
        """Returns True if the state has changed since the last time this method was called"""
        state_change_marker = self._redis_client.get(
            f"{self._workcell_prefix}:state_changed"
        )
        if state_change_marker != self.state_change_marker:
            self.state_change_marker = state_change_marker
            return True
        return False

    # *Workcell Methods
    def get_workcell(self) -> WorkcellDefinition:
        """
        Returns the current workcell as a Workcell object
        """
        return WorkcellDefinition.model_validate(self._workcell.to_dict())

    def set_workcell(self, workcell: WorkcellDefinition) -> None:
        """
        Sets the active workcell
        """
        self._workcell.update(**workcell.model_dump(mode="json"))

    def clear_workcell(self) -> None:
        """
        Empty the workcell definition
        """
        self._workcell.clear()

    def get_workcell_id(self) -> str:
        """
        Returns the workcell ID
        """
        wc_id = self._redis_client.get(f"{self._workcell_prefix}:workcell_id")
        if wc_id is None:
            self._redis_client.set(
                f"{self._workcell_prefix}:workcell_id", new_ulid_str()
            )
            wc_id = self._redis_client.get(f"{self._workcell_prefix}:workcell_id")
        return wc_id

    # *Workflow Methods
    def get_workflow(self, workflow_id: Union[str, str]) -> Workflow:
        """
        Returns a workflow by ID
        """
        return Workflow.model_validate(self._workflows[str(workflow_id)])

    def get_all_workflows(self) -> dict[str, Workflow]:
        """
        Returns all workflow runs
        """
        valid_workflows = {}
        for workflow_id, workflow in self._workflows.to_dict().items():
            try:
                valid_workflows[str(workflow_id)] = Workflow.model_validate(workflow)
            except ValidationError:
                continue
        return valid_workflows

    def set_workflow(self, wf: Workflow) -> None:
        """
        Sets a workflow by ID
        """
        if isinstance(wf, Workflow):
            wf_dump = wf.model_dump(mode="json")
        else:
            wf_dump = Workflow.model_validate(wf).model_dump(mode="json")
        self._workflows[str(wf_dump["workflow_id"])] = wf_dump
        self.mark_state_changed()

    def set_workflow_quiet(self, wf: Workflow) -> None:
        """
        Sets a workflow by ID
        """
        if isinstance(wf, Workflow):
            wf_dump = wf.model_dump(mode="json")
        else:
            wf_dump = Workflow.model_validate(wf).model_dump(mode="json")
        self._workflows[str(wf_dump["workflow_id"])] = wf_dump

    def delete_workflow(self, workflow_id: Union[str, str]) -> None:
        """
        Deletes a workflow by ID
        """
        del self._workflows[str(workflow_id)]
        self.mark_state_changed()

    def update_workflow(
        self, workflow_id: str, func: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> None:
        """
        Updates the state of a workflow.
        """
        self.set_workflow(func(self.get_workflow(workflow_id), *args, **kwargs))

    def get_node(self, node_name: str) -> Node:
        """
        Returns a node by name
        """
        return Node.model_validate(self._nodes[node_name])

    def get_all_nodes(self) -> dict[str, Node]:
        """
        Returns all nodes
        """
        valid_nodes = {}
        for node_name, node in self._nodes.to_dict().items():
            try:
                valid_nodes[str(node_name)] = Node.model_validate(node)
            except ValidationError:
                continue
        return valid_nodes

    def set_node(
        self, node_name: str, node: Union[Node, NodeDefinition, dict[str, Any]]
    ) -> None:
        """
        Sets a node by name
        """
        if isinstance(node, Node):
            node_dump = node.model_dump(mode="json")
        elif isinstance(node, NodeDefinition):
            node_dump = Node.model_validate(node, from_attributes=True).model_dump(
                mode="json"
            )
        else:
            node_dump = Node.model_validate(node).model_dump(mode="json")
        self._nodes[node_name] = node_dump
        self.mark_state_changed()

    def delete_node(self, node_name: str) -> None:
        """
        Deletes a node by name
        """
        del self._nodes[node_name]
        self.mark_state_changed()

    def update_node(
        self, node_name: str, func: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> None:
        """
        Updates the state of a node.
        """
        self.set_node(node_name, func(self.get_node(node_name), *args, **kwargs))



"""Automated pytest unit tests for the madsci workcell manager's REST server."""

import pytest
from fastapi.testclient import TestClient
from madsci.common.types.node_types import Node
from madsci.common.types.workcell_types import WorkcellDefinition
from madsci.common.types.workflow_types import Workflow, WorkflowDefinition
from madsci.workcell_manager.workcell_server import create_workcell_server
from pydantic import AnyUrl
from pytest_mock_resources import RedisConfig, create_redis_fixture
from redis import Redis


# Create a Redis server fixture for testing
@pytest.fixture(scope="session")
def pmr_redis_config() -> RedisConfig:
    """Configure the Redis server."""
    return RedisConfig(image="redis:7.4")


redis_server = create_redis_fixture()


@pytest.fixture
def workcell() -> WorkcellDefinition:
    """Fixture for creating a WorkcellDefinition."""
    # TODO: Add node(s) to this workcell for testing purposes
    return WorkcellDefinition(workcell_name="Test Workcell")


@pytest.fixture
def test_client(workcell: WorkcellDefinition, redis_server: Redis) -> TestClient:
    """Workcell Server Test Client Fixture"""
    app = create_workcell_server(workcell, redis_server, start_engine=False)
    return TestClient(app)


def test_get_workcell(test_client: TestClient) -> None:
    """Test the /definition endpoint."""
    with test_client as client:
        response = client.get("/definition")
        assert response.status_code == 200
        WorkcellDefinition.model_validate(response.json())


def test_get_nodes(test_client: TestClient) -> None:
    """Test the /nodes endpoint."""
    with test_client as client:
        response = client.get("/nodes")
        assert response.status_code == 200
        assert isinstance(response.json(), dict)


def test_add_node(test_client: TestClient) -> None:
    """Test adding a node to the workcell."""
    with test_client as client:
        node_name = "test_node"
        node_url = "http://localhost:8000"
        response = client.post(
            "/nodes/add_node",
            params={
                "node_name": node_name,
                "node_url": node_url,
                "node_description": "A Node",
                "permanent": False,
            },
        )
        assert response.status_code == 200
        node = Node.model_validate(response.json())
        assert node.node_url == AnyUrl(node_url)


def test_send_admin_command(test_client: TestClient) -> None:
    """Test sending an admin command to all nodes."""
    with test_client as client:
        response = client.get("/admin/reset")
        assert response.status_code == 200
        assert isinstance(response.json(), list)


def test_get_all_workflows(test_client: TestClient) -> None:
    """Test the /workflows endpoint."""
    with test_client as client:
        response = client.get("/workflows")
        assert response.status_code == 200
        assert isinstance(response.json(), dict)


def test_start_workflow(test_client: TestClient) -> None:
    """Test starting a new workflow."""
    with test_client as client:
        workflow_def = WorkflowDefinition(name="Test Workflow")
        response = client.post(
            "/workflows/start",
            data={"workflow": workflow_def.model_dump_json(), "validate_only": False},
        )
        assert response.status_code == 200
        workflow = Workflow.model_validate(response.json())
        assert workflow.name == workflow_def.name
        response = client.get(f"/workflows/{workflow.workflow_id}")
        assert response.status_code == 200
        workflow = Workflow.model_validate(response.json())
        assert workflow.name == workflow_def.name


def test_pause_and_resume_workflow(test_client: TestClient) -> None:
    """Test pausing and resuming a workflow."""
    with test_client as client:
        workflow_def = WorkflowDefinition(name="Test Workflow")
        response = client.post(
            "/workflows/start",
            data={"workflow": workflow_def.model_dump_json(), "validate_only": False},
        )
        workflow = Workflow.model_validate(response.json())
        response = client.post(f"/workflows/pause/{workflow.workflow_id}")
        assert response.status_code == 200
        workflow = Workflow.model_validate(response.json())
        assert workflow.paused is True
        response = test_client.post(f"/workflows/resume/{workflow.workflow_id}")
        assert response.status_code == 200
        workflow = Workflow.model_validate(response.json())
        assert workflow.paused is False


def test_cancel_workflow(test_client: TestClient) -> None:
    """Test canceling and resubmitting a workflow."""
    with test_client as client:
        workflow_def = WorkflowDefinition(name="Test Workflow")
        response = client.post(
            "/workflows/start",
            data={"workflow": workflow_def.model_dump_json(), "validate_only": False},
        )
        workflow = Workflow.model_validate(response.json())
        response = test_client.post(f"/workflows/cancel/{workflow.workflow_id}")
        assert response.status_code == 200
        workflow = Workflow.model_validate(response.json())
        assert workflow.status == "cancelled"
        response = test_client.post(f"/workflows/resubmit/{workflow.workflow_id}")
        assert response.status_code == 200
        new_workflow = Workflow.model_validate(response.json())
        assert workflow.workflow_id != new_workflow.workflow_id


def test_retry_workflow(test_client: TestClient) -> None:
    """Test retrying a workflow."""
    with test_client as client:
        workflow_def = WorkflowDefinition(name="Test Workflow")
        response = client.post(
            "/workflows/start",
            data={"workflow": workflow_def.model_dump_json(), "validate_only": False},
        )
        workflow = Workflow.model_validate(response.json())
        response = test_client.post(f"/workflows/cancel/{workflow.workflow_id}")
        assert response.status_code == 200
        workflow = Workflow.model_validate(response.json())
        assert workflow.status == "cancelled"
        response = test_client.post(
            f"/workflows/retry/{workflow.workflow_id}", params={"index": 0}
        )
        assert response.status_code == 200
        new_workflow = Workflow.model_validate(response.json())
        assert workflow.workflow_id == new_workflow.workflow_id
        assert new_workflow.status == "queued"



"""Functions for checking conditions on a step"""

from madsci.common.types.condition_types import (
    NoResourceInLocationCondition,
    ResourceInLocationCondition,
)
from madsci.common.types.resource_types import Container
from madsci.common.types.step_types import Step
from madsci.common.types.workflow_types import Workflow
from madsci.workcell_manager.schedulers.default_scheduler import Scheduler


def evaluate_condition_checks(step: Step, wf: Workflow, scheduler: Scheduler) -> None:
    """Check if the specified conditions for the step are met"""
    for condition in step.conditions:
        if isinstance(condition, ResourceInLocationCondition):
            evaluate_resource_in_location_condition(condition, wf, scheduler)
        elif isinstance(condition, NoResourceInLocationCondition):
            evaluate_no_resource_in_location_condition(condition, wf, scheduler)
        else:
            raise ValueError(f"Unknown condition type {condition.condition_type}")


def evaluate_resource_in_location_condition(
    condition: ResourceInLocationCondition, wf: Workflow, scheduler: Scheduler
) -> None:
    """Check if a resource is present in a specified location"""
    location = scheduler.workcell.locations[condition.location]
    if location.resource_id is None:
        wf.scheduler_metadata.ready_to_run = False
        wf.scheduler_metadata.reasons.append(
            f"Location {location.name} cannot provide resource presence information."
        )
    elif scheduler.resource_client is None:
        wf.scheduler_metadata.ready_to_run = False
        wf.scheduler_metadata.reasons.append("Resource client is not available.")
    else:
        container = scheduler.resource_client.get_resource(location.resource_id)
        if not isinstance(container, Container):
            wf.scheduler_metadata.ready_to_run = False
            wf.scheduler_metadata.reasons.append(
                f"Resource {container.resource_id} is not a container."
            )
            return
        if len(container.children) == 0:
            wf.scheduler_metadata.ready_to_run = False
            wf.scheduler_metadata.reasons.append(
                f"Resource {container.resource_id} is empty."
            )
        if condition.key is not None and container.get_child(condition.key) is None:
            wf.scheduler_metadata.ready_to_run = False
            wf.scheduler_metadata.reasons.append(
                f"Resource {container.resource_id} does not contain a child with key {condition.key}."
            )


def evaluate_no_resource_in_location_condition(
    condition: NoResourceInLocationCondition, wf: Workflow, scheduler: Scheduler
) -> None:
    """Check if a resource is not present in a specified location"""
    location = scheduler.workcell.locations[condition.location]
    if location.resource_id is None:
        wf.scheduler_metadata.ready_to_run = False
        wf.scheduler_metadata.reasons.append(
            f"Location {location.name} cannot provide resource presence information."
        )
    elif scheduler.resource_client is None:
        wf.scheduler_metadata.ready_to_run = False
        wf.scheduler_metadata.reasons.append("Resource client is not available.")
    else:
        container = scheduler.resource_client.get_resource(location.resource_id)
        if not isinstance(container, Container):
            wf.scheduler_metadata.ready_to_run = False
            wf.scheduler_metadata.reasons.append(
                f"Resource {container.resource_id} is not a container."
            )
            return
        if len(container.children) > 0:
            wf.scheduler_metadata.ready_to_run = False
            wf.scheduler_metadata.reasons.append(
                f"Resource {container.resource_id} is not empty."
            )
        if condition.key is not None and container.get_child(condition.key) is not None:
            wf.scheduler_metadata.ready_to_run = False
            wf.scheduler_metadata.reasons.append(
                f"Resource {container.resource_id} contains a child with key {condition.key} ({container.get_child(condition.key).resource_id})."
            )



"""
Engine Class and associated helpers and data
"""

import importlib
import time
import traceback
from datetime import datetime
from typing import Optional

from madsci.client.data_client import DataClient
from madsci.client.event_client import default_logger
from madsci.client.node.abstract_node_client import AbstractNodeClient
from madsci.common.types.action_types import ActionRequest, ActionResult
from madsci.common.types.datapoint_types import FileDataPoint, ValueDataPoint
from madsci.common.types.node_types import Node
from madsci.common.types.step_types import Step
from madsci.common.types.workcell_types import WorkcellDefinition
from madsci.common.types.workflow_types import Workflow, WorkflowStatus
from madsci.common.utils import threaded_daemon
from madsci.workcell_manager.redis_handler import WorkcellRedisHandler
from madsci.workcell_manager.workcell_utils import (
    find_node_client,
    initialize_workcell,
    update_active_nodes,
)
from madsci.workcell_manager.workflow_utils import cancel_active_workflows


class Engine:
    """
    Handles scheduling workflows and executing steps on the workcell.
    Pops incoming workflows off a redis-based queue and executes them.
    """

    def __init__(
        self,
        workcell_manager_definition: WorkcellDefinition,
        state_handler: WorkcellRedisHandler,
    ) -> None:
        """Initialize the scheduler."""
        state_handler.clear_state(
            clear_workflows=workcell_manager_definition.config.clear_workflows
        )
        self.definition = workcell_manager_definition
        self.state_handler = state_handler
        cancel_active_workflows(state_handler)
        scheduler_module = importlib.import_module(self.definition.config.scheduler)
        self.scheduler = scheduler_module.Scheduler(self.definition, self.state_handler)
        self.data_client = DataClient(self.definition.config.data_server_url)
        with state_handler.wc_state_lock():
            initialize_workcell(state_handler)
        time.sleep(workcell_manager_definition.config.cold_start_delay)
        default_logger.log_info("Engine initialized, waiting for workflows...")
        # TODO send event

    def spin(self) -> None:
        """
        Continuously loop, updating module states every Config.update_interval seconds.
        If the state of the workcell has changed, update the active modules and run the scheduler.
        """
        update_active_nodes(self.state_handler)
        node_tick = time.time()
        scheduler_tick = time.time()
        heartbeat = time.time()
        while True and not self.state_handler.shutdown:
            try:
                if time.time() - heartbeat > self.definition.config.heartbeat_interval:
                    heartbeat = time.time()
                    default_logger.log_info(f"Heartbeat: {time.time()}")
                if (
                    time.time() - node_tick
                    > self.definition.config.node_update_interval
                    or self.state_handler.has_state_changed()
                ):
                    if not self.state_handler.paused:
                        update_active_nodes(self.state_handler)
                    node_tick = time.time()
                if (
                    time.time() - scheduler_tick
                    > self.definition.config.scheduler_update_interval
                ):
                    with self.state_handler.wc_state_lock():
                        self.scheduler.run_iteration(
                            workflows=self.state_handler.get_all_workflows()
                        )
                        self.run_next_step()
                        scheduler_tick = time.time()
            except Exception:
                traceback.print_exc()
                default_logger.log_error(
                    f"Error in engine loop, waiting {self.definition.config.node_update_interval} seconds before trying again."
                )
                time.sleep(self.definition.config.node_update_interval)

    @threaded_daemon
    def start_engine_thread(self) -> None:
        """Spins the engine in its own thread"""
        self.spin()

    def run_next_step(self) -> None:
        """Runs the next step in the workflow with the highest priority"""
        workflows = self.state_handler.get_all_workflows()
        ready_workflows = filter(
            lambda wf: wf.scheduler_metadata.ready_to_run, workflows.values()
        )
        sorted_ready_workflows = sorted(
            ready_workflows, key=lambda wf: wf.scheduler_metadata.priority, reverse=True
        )
        if len(sorted_ready_workflows) > 0:
            next_wf = sorted_ready_workflows[0]
            next_wf.status = WorkflowStatus.RUNNING
            self.state_handler.set_workflow(next_wf)
            self.run_step(next_wf.workflow_id, next_wf.steps[next_wf.step_index])

    def query_action_result(
        self,
        node: Node,
        client: AbstractNodeClient,
        request: ActionRequest,
        response: Optional[ActionResult] = None,
    ) -> ActionResult:
        """retry an action if it fails"""
        if node.info.capabilities.get_action_result:
            while response is None or response.status not in [
                "not_ready",
                "succeeded",
                "failed",
            ]:
                try:
                    response = client.get_action_result(request.action_id)
                    time.sleep(5)
                except Exception:
                    time.sleep(5)
            return response
        return response

    @threaded_daemon
    def run_step(self, workflow_id: str, step: Step) -> None:
        """Run a step in a seperate thread"""
        with self.state_handler.wc_state_lock():
            wf = self.state_handler.get_workflow(workflow_id)
            wf.steps[wf.step_index].start_time = datetime.now()
            if wf.step_index == 0:
                wf.start_time = datetime.now()
            self.state_handler.set_workflow(wf)
        node = self.state_handler.get_node(step.node)
        client = find_node_client(node.node_url)
        response = None
        try:
            request = ActionRequest(
                action_name=step.action,
                args={**step.args, **step.locations},
                files=step.files,
            )
            response = client.send_action(request)
        except Exception:
            default_logger.log_info("request had exception")
        finally:
            response = self.query_action_result(node, client, request, response)
        if response is None:
            response = request.failed()
        response = self.handle_data_and_files(response)
        with self.state_handler.wc_state_lock():
            wf = self.state_handler.get_workflow(workflow_id)
            if response.status in ["succeeded", "failed"]:
                wf.steps[wf.step_index].status = response.status
                wf.steps[wf.step_index].results[response.action_id] = response
                wf.steps[wf.step_index].end_time = datetime.now()
                if response.status == "succeeded":
                    new_index = wf.step_index + 1
                    if new_index == len(wf.steps):
                        wf.status = WorkflowStatus.COMPLETED
                        wf.end_time = datetime.now()
                    else:
                        wf.step_index = new_index
                        if wf.status == WorkflowStatus.RUNNING:
                            wf.status = WorkflowStatus.IN_PROGRESS
                if response.status == "failed":
                    wf.status = WorkflowStatus.FAILED
                    wf.end_time = datetime.now()
            self.state_handler.set_workflow(wf)

    def handle_data_and_files(
        self, step: Step, wf: Workflow, response: ActionResult
    ) -> ActionResult:
        """create and save datapoints for data returned from step"""
        labeled_data = {}
        if response.data:
            for data_key in response.data:
                if step.data_labels is not None and data_key in step.data_labels:
                    label = step.data_labels[data_key]
                else:
                    label = data_key
                datapoint = ValueDataPoint(
                    label=label,
                    step_id=step.id,
                    workflow_id=wf.workflow_id,
                    experiment_id=wf.experiment_id,
                    value=response.data[data_key],
                )
                self.data_client.submit_datapoint(datapoint)
                labeled_data[label] = datapoint.id
        if response.files:
            for file_key in response.files:
                if step.data_labels is not None and file_key in step.data_labels:
                    label = step.data_labels[file_key]
                else:
                    label = file_key
                datapoint = FileDataPoint(
                    step_id=step.id,
                    workflow_id=wf.workflow_id,
                    experiment_id=wf.experiment_id,
                    label=label,
                    path=str(response.files[file_key]),
                )
                self.data_client.submit_datapoint(datapoint, datapoint.path)
                labeled_data[label] = datapoint.id
        response.data = labeled_data
        return response



"""utility functions for the workcell"""

import concurrent
import warnings
from typing import Optional

import requests
from madsci.client.node import NODE_CLIENT_MAP, AbstractNodeClient
from madsci.common.types.node_types import Node, NodeDefinition
from madsci.common.types.workcell_types import WorkcellDefinition, WorkcellLink
from madsci.workcell_manager.redis_handler import WorkcellRedisHandler
from pydantic import AnyUrl


def resolve_workcell_link(workcell_link: WorkcellLink) -> WorkcellDefinition:
    """Resolves the workcell link to a workcell definition"""
    if workcell_link.definition is not None:
        return workcell_link.definition
    if workcell_link.url is not None:
        return WorkcellDefinition.model_validate(
            requests.get(workcell_link.url, timeout=10).json()
        )
    if workcell_link.path is not None:
        return WorkcellDefinition.from_yaml(workcell_link.path)
    return workcell_link.expanduser()


def initialize_workcell(
    state_manager: WorkcellRedisHandler, workcell: Optional[WorkcellDefinition] = None
) -> None:
    """
    Initializes the state of the workcell from the workcell definition.
    """

    if not workcell:
        workcell = state_manager.get_workcell()
    initialize_workcell_nodes(workcell, state_manager)
    initialize_workcell_resources(workcell)


def initialize_workcell_nodes(
    workcell: WorkcellDefinition, state_manager: WorkcellRedisHandler
) -> None:
    """create the nodes for the given workcell"""
    for key, value in workcell.nodes.items():
        if type(value) is NodeDefinition:
            node = Node(node_url=value.node_url)
        elif type(value) is AnyUrl or type(value) is str:
            node = Node(node_url=AnyUrl(value))
        state_manager.set_node(key, node)


def initialize_workcell_resources(workcell: WorkcellDefinition) -> None:
    """create the resources for a given workcell definition"""


def find_node_client(url: str) -> AbstractNodeClient:
    """finds the right client for the node url provided"""
    for client in NODE_CLIENT_MAP.values():
        if client.validate_url(url):
            return client(url)
    for client in AbstractNodeClient.__subclasses__():
        if client.validate_url(url):
            return client(url)
    return None


def update_active_nodes(state_manager: WorkcellRedisHandler) -> None:
    """Update all active nodes in the workcell."""
    with concurrent.futures.ThreadPoolExecutor() as executor:
        node_futures = []
        for node_name, node in state_manager.get_all_nodes().items():
            node_future = executor.submit(update_node, node_name, node, state_manager)
            node_futures.append(node_future)

        # Wait for all node updates to complete
        concurrent.futures.wait(node_futures)


def update_node(
    node_name: str, node: Node, state_manager: WorkcellRedisHandler
) -> None:
    """Update a single node's state and about information."""
    try:
        old_status = node.status
        old_info = node.info
        client = find_node_client(node.node_url)
        node.status = client.get_status()
        node.info = client.get_info()
        node.state = client.get_state()
        if old_status != node.status or old_info != node.info:
            with state_manager.wc_state_lock():
                state_manager.set_node(node_name, node)
    except Exception:
        warnings.warn(  # TODO: Replace with event logger
            message=f"Unable to update node {node_name}",
            category=UserWarning,
            stacklevel=1,
        )



"""the abstract class for schedulers"""

from typing import ClassVar, Optional

from madsci.client.event_client import EventClient
from madsci.client.resource_client import ResourceClient
from madsci.common.types.event_types import Event
from madsci.common.types.workcell_types import WorkcellDefinition
from madsci.common.types.workflow_types import Workflow
from madsci.workcell_manager.redis_handler import WorkcellRedisHandler


def send_event(test: Event) -> None:  # TODO: remove placeholder
    """send an event to the server"""


class AbstractScheduler:
    """abstract definition of a scheduler"""

    workcell_definition: ClassVar[WorkcellDefinition]
    running: bool
    state_handler: WorkcellRedisHandler
    logger: Optional[EventClient]
    resource_client: Optional[ResourceClient]

    def __init__(
        self,
        workcell_definition: WorkcellDefinition,
        state_handler: WorkcellRedisHandler,
    ) -> "AbstractScheduler":
        """sets the state handler and workcell definition"""
        self.state_handler = state_handler
        self.workcell_definition = workcell_definition
        self.running = True
        self.logger = EventClient(
            config=self.workcell_definition.config.event_client_config
        )
        if self.workcell_definition.config.resource_server_url is not None:
            self.resource_client = ResourceClient(
                url=self.workcell_definition.config.resource_server_url
            )
        else:
            self.resource_client = None

    def run_iteration(self, workflows: list[Workflow]) -> None:
        """Run an iteration of the scheduler"""



"""Default MADSci Workcell scheduler"""

import traceback

from madsci.common.types.step_types import Step
from madsci.common.types.workflow_types import Workflow
from madsci.workcell_manager.condition_checks import evaluate_condition_checks
from madsci.workcell_manager.schedulers.scheduler import AbstractScheduler


class Scheduler(AbstractScheduler):
    """The main class that handles checking whether steps are ready to run and assigning priority"""

    def run_iteration(self, workflows: list[Workflow]) -> None:
        """Run an iteration of the scheduling algorithm and markup the scheduler metadata for each workflow"""
        priority = 0
        workflows: list[Workflow] = sorted(
            self.state_handler.get_all_workflows().values(),
            key=lambda item: item.submitted_time,
        )
        for wf in workflows:
            try:
                if wf.step_index < len(wf.steps):
                    step = wf.steps[wf.step_index]
                    wf.scheduler_metadata.ready_to_run = True
                    wf.scheduler_metadata.reasons = []
                    if wf.paused:
                        wf.scheduler_metadata.ready_to_run = False
                        wf.scheduler_metadata.reasons.append("Workflow is paused")
                    if wf.status not in ["queued", "in_progress"]:
                        wf.scheduler_metadata.ready_to_run = False
                        wf.scheduler_metadata.reasons.append(
                            "Workflow is not queued or in progress"
                        )
                    self.location_checks(step, wf)
                    self.resource_checks(step, wf)
                    self.node_checks(step, wf)
                    evaluate_condition_checks(
                        step, wf, self.workcell_definition, self.state_handler
                    )
                    wf.scheduler_metadata.priority = priority
                    priority -= 1
            except Exception as e:
                self.logger.log_error(
                    f"Error in scheduler while evaluating workflow {wf.workflow_id}: {traceback.format_exc()}"
                )
                wf.scheduler_metadata.ready_to_run = False
                wf.scheduler_metadata.reasons.append(f"Exception in scheduler: {e}")
            finally:
                self.state_handler.set_workflow_quiet(wf)

    def location_checks(self, step: Step, wf: Workflow) -> None:
        """Check if the location(s) for the step are ready"""
        for location in step.locations.values():
            if location.resource_id is not None and self.resource_client is not None:
                self.resource_client.get_resource(location.resource_id)
                # TODO: what do we do with the location_resource?
            if location.reservation is not None:
                wf.scheduler_metadata.ready_to_run = False
                wf.scheduler_metadata.reasons.append(
                    f"Location {location.location_id} is reserved by {location.reservation.owned_by.model_dump(mode='json', exclude_none=True)}"
                )

    def resource_checks(self, step: Step, wf: Workflow) -> None:
        """Check if the resources for the step are ready TODO: actually check"""

    def node_checks(self, step: Step, wf: Workflow) -> None:
        """Check if the node used in the step currently has a "ready" status"""
        node = self.state_handler.get_node(step.node)
        if node is None:
            wf.scheduler_metadata.ready_to_run = False
            wf.scheduler_metadata.reasons.append(f"Node {step.node} not found")
        if not node.status.ready:
            wf.scheduler_metadata.ready_to_run = False
            wf.scheduler_metadata.reasons.append(
                f"Node {step.node} not ready: {node.status.description}"
            )
        if node.reservation is not None and node.reservation.check(wf.ownership_info):
            wf.scheduler_metadata.ready_to_run = False
            wf.scheduler_metadata.reasons.append(
                f"Node {step.node} is reserved by {node.reservation.owned_by.model_dump(mode='json', exclude_none=True)}"
            )



"""MADSci Workcell Manager Server."""

import json
import traceback
from contextlib import asynccontextmanager
from datetime import datetime
from typing import Annotated, Any, Optional, Union

from fastapi import FastAPI, Form, HTTPException, UploadFile
from madsci.common.types.action_types import ActionStatus
from madsci.common.types.auth_types import OwnershipInfo
from madsci.common.types.base_types import new_ulid_str
from madsci.common.types.node_types import Node, NodeDefinition
from madsci.common.types.workcell_types import WorkcellDefinition
from madsci.common.types.workflow_types import (
    Workflow,
    WorkflowDefinition,
    WorkflowStatus,
)
from madsci.workcell_manager.redis_handler import WorkcellRedisHandler
from madsci.workcell_manager.workcell_engine import Engine
from madsci.workcell_manager.workcell_utils import find_node_client
from madsci.workcell_manager.workflow_utils import (
    copy_workflow_files,
    create_workflow,
    save_workflow_files,
)


def create_workcell_server(  # noqa: C901, PLR0915
    workcell: WorkcellDefinition,
    redis_connection: Optional[Any] = None,
    start_engine: bool = True,
) -> FastAPI:
    """Creates a Workcell Manager's REST server."""

    state_handler = WorkcellRedisHandler(workcell, redis_connection=redis_connection)

    @asynccontextmanager
    async def lifespan(app: FastAPI):  # noqa: ANN202, ARG001
        """Start the REST server and initialize the state handler and engine"""
        state_handler.set_workcell(workcell)
        if start_engine:
            engine = Engine(workcell, state_handler)
            engine.start_engine_thread()
        yield

    app = FastAPI(lifespan=lifespan)

    @app.get("/definition")
    @app.get("/workcell")
    def get_workcell() -> WorkcellDefinition:
        """Get the currently running workcell."""
        return state_handler.get_workcell()

    @app.get("/state")
    def get_state() -> dict:
        """Get the current state of the workcell."""
        return state_handler.get_state()

    @app.get("/nodes")
    def get_nodes() -> dict[str, Node]:
        """Get info on the nodes in the workcell."""
        return state_handler.get_all_nodes()

    @app.get("/nodes/{node_name}")
    def get_node(node_name: str) -> Union[Node, str]:
        """Get information about about a specific node."""
        try:
            node = state_handler.get_node(node_name)
        except Exception:
            return "Node not found!"
        return node

    @app.post("/nodes/add_node")
    def add_node(
        node_name: str,
        node_url: str,
        node_description: str = "A Node",
        permanent: bool = False,
    ) -> Union[Node, str]:
        """Add a node to the workcell's node list"""
        if node_name in state_handler.get_all_nodes():
            return "Node name exists, node names must be unique!"
        node = Node(node_url=node_url)
        state_handler.set_node(node_name, node)
        if permanent:
            workcell.nodes[node_name] = NodeDefinition(
                node_name=node_name,
                node_url=node_url,
                node_description=node_description,
            )
            workcell.to_yaml(workcell._definition_path)
        return state_handler.get_node(node_name)

    @app.get("/admin/{command}")
    def send_admin_command(command: str) -> list:
        """Send an admin command to all capable nodes."""
        responses = []
        for node in state_handler.get_all_nodes().values():
            if command in node.info.capabilities.admin_commands:
                client = find_node_client(node.node_url)
                response = client.send_admin_command(command)
                responses.append(response)
        return responses

    @app.get("/admin/{command}/{node}")
    def send_admin_command_to_node(command: str, node: str) -> list:
        """Send admin command to a node."""
        responses = []
        node = state_handler.get_node(node)
        if command in node.info.capabilities.admin_commands:
            client = find_node_client(node.node_url)
            response = client.send_admin_command(command)
            responses.append(response)
        return responses

    @app.get("/workflows")
    def get_all_workflows() -> dict[str, Workflow]:
        """get all workflows."""
        return state_handler.get_all_workflows()

    @app.get("/workflows/{workflow_id}")
    def get_workflow(workflow_id: str) -> Workflow:
        """Get info on a specific workflow."""
        return state_handler.get_workflow(workflow_id)

    @app.post("/workflows/pause/{workflow_id}")
    def pause_workflow(workflow_id: str) -> Workflow:
        """Pause a specific workflow."""
        with state_handler.wc_state_lock():
            wf = state_handler.get_workflow(workflow_id)
            if wf.status in ["running", "in_progress", "queued"]:
                if wf.status == "running":
                    send_admin_command_to_node("pause", wf.steps[wf.step_index].node)
                    wf.steps[wf.step_index] = ActionStatus.PAUSED
                wf.paused = True
                state_handler.set_workflow(wf)

        return state_handler.get_workflow(workflow_id)

    @app.post("/workflows/resume/{workflow_id}")
    def resume_workflow(workflow_id: str) -> Workflow:
        """Resume a paused workflow."""
        with state_handler.wc_state_lock():
            wf = state_handler.get_workflow(workflow_id)
            if wf.paused:
                if wf.status == "running":
                    send_admin_command_to_node("resume", wf.steps[wf.step_index].node)
                    wf.steps[wf.step_index] = ActionStatus.RUNNING
                wf.paused = False
                state_handler.set_workflow(wf)
        return state_handler.get_workflow(workflow_id)

    @app.post("/workflows/cancel/{workflow_id}")
    def cancel_workflow(workflow_id: str) -> Workflow:
        """Cancel a specific workflow."""
        with state_handler.wc_state_lock():
            wf = state_handler.get_workflow(workflow_id)
            if wf.status == "running":
                send_admin_command_to_node("stop", wf.steps[wf.step_index].node)
                wf.steps[wf.step_index] = ActionStatus.CANCELLED
            wf.status = WorkflowStatus.CANCELLED
            state_handler.set_workflow(wf)
        return state_handler.get_workflow(workflow_id)

    @app.post("/workflows/resubmit/{workflow_id}")
    def resubmit_workflow(workflow_id: str) -> Workflow:
        """resubmit a previous workflow as a new workflow."""
        with state_handler.wc_state_lock():
            wf = state_handler.get_workflow(workflow_id)
            wf.workflow_id = new_ulid_str()
            wf.step_index = 0
            wf.start_time = None
            wf.end_time = None
            wf.submitted_time = datetime.now()
            for step in wf.steps:
                step.step_id = new_ulid_str()
                step.start_time = None
                step.end_time = None
                step.status = ActionStatus.NOT_STARTED
            copy_workflow_files(
                old_id=workflow_id,
                workflow=wf,
                working_directory=workcell.workcell_directory,
            )
            state_handler.set_workflow(wf)
        return state_handler.get_workflow(wf.workflow_id)

    @app.post("/workflows/retry/{workflow_id}")
    def retry_workflow(workflow_id: str, index: int = -1) -> Workflow:
        """Retry an existing workflow from a specific step."""
        with state_handler.wc_state_lock():
            wf = state_handler.get_workflow(workflow_id)
            if wf.status in [
                WorkflowStatus.COMPLETED,
                WorkflowStatus.FAILED,
                WorkflowStatus.CANCELLED,
            ]:
                if index >= 0:
                    wf.step_index = index
                wf.status = WorkflowStatus.QUEUED
                state_handler.set_workflow(wf)
        return state_handler.get_workflow(workflow_id)

    @app.post("/workflows/start")
    async def start_workflow(
        workflow: Annotated[str, Form()],
        ownership_info: Annotated[Optional[str], Form()] = None,
        parameters: Annotated[Optional[str], Form()] = None,
        validate_only: Annotated[Optional[bool], Form()] = False,
        files: list[UploadFile] = [],
    ) -> Workflow:
        """
        parses the payload and workflow files, and then pushes a workflow job onto the redis queue

        Parameters
        ----------
        workflow: UploadFile
        - The workflow yaml file
        parameters: Optional[Dict[str, Any]] = {}
        - Dynamic values to insert into the workflow file
        ownership_info: Optional[OwnershipInfo]
        - Information about the experiments, users, etc. that own this workflow
        simulate: bool
        - whether to use real robots or not
        validate_only: bool
        - whether to validate the workflow without queueing it

        Returns
        -------
        response: Workflow
        - a workflow run object for the requested run_id
        """
        try:
            wf_def = WorkflowDefinition.model_validate_json(workflow)
        except Exception as e:
            traceback.print_exc()
            raise HTTPException(status_code=422, detail=str(e)) from e

        if ownership_info is None:
            ownership_info = OwnershipInfo()
        else:
            ownership_info = OwnershipInfo.model_validate_json(ownership_info)

        if parameters is None:
            parameters = {}
        else:
            parameters = json.loads(parameters)
            if not isinstance(parameters, dict) or not all(
                isinstance(k, str) for k in parameters
            ):
                raise HTTPException(
                    status_code=400,
                    detail="Parameters must be a dictionary with string keys",
                )
        workcell = state_handler.get_workcell()

        wf = create_workflow(
            workflow_def=wf_def,
            workcell=workcell,
            ownership_info=ownership_info,
            parameters=parameters,
            state_handler=state_handler,
        )

        if not validate_only:
            wf = save_workflow_files(
                working_directory=workcell.workcell_directory,
                workflow=wf,
                files=files,
            )
            with state_handler.wc_state_lock():
                state_handler.set_workflow(wf)
        return wf

    return app


if __name__ == "__main__":
    import uvicorn

    workcell = None
    workcell = WorkcellDefinition.load_model(require_unique=True)
    if workcell is None:
        raise ValueError(
            "No workcell manager definition found, please specify a path with --definition, or add it to your lab definition's 'managers' section"
        )
    app = create_workcell_server(workcell)
    uvicorn.run(
        app,
        host=workcell.config.host,
        port=workcell.config.port,
    )
